# Python design pattern pou parcourir d'une base de donnée

![category](développement)
![tag](designpattern)
![tag](python)


En utilisant les design partenaire d'arbre et de filtre, on peut facilement gérer les tables d'une base de données
ainsi que les relations entre table.

Dans notre exemple nous utilisons une table contenant deux tables (user, email).

Nous allons créer les objets GenericList, GenericModel et les objets dérivés spécifiques Users, User, Emails, Email

L'utilisation et le parcours des tables devien alors très faciles

```python
import sqlite3
import os
import collections
from copy import deepcopy

class returnStillNothing(str):

    def __getattr__(self, id):
        return returnStillNothing()

    def __getitem__(self, i):
        return returnStillNothing()

    def __call__(self, *args, **kw):
        return returnStillNothing()



class GenericModel():

    def __init__(self, **kw):
        for key in kw:
            self.__setattr__(key, kw[key])

    def __getattr__(self, id):
        return returnStillNothing()


class GenericList(collections.UserList):

    def __init__(self):
        collections.UserList.__init__(self)

    def __getattr__(self, id):
        if len(self) > 0:
            return getattr(self[0], id)
        return returnStillNothing()

    def _copy_clean(self):
        obj = deepcopy(self)
        obj.clear()
        return obj

    def filter(self, *args, **kw):
        if len(args) > 0:
            if not callable(args[0]):
                raise ValueError("%s is not callable" % args[0])
            obj = self._copy_clean()
            for elt in self:
                if args[0](elt) == True:
                    obj.append(elt)
            if len(args[1:]) > 0:
                return obj.filter(*args[1:], **kw)
            return obj.filter(**kw)
        if len(kw) > 0:
            key = list(kw.keys())[0]
            obj = self._copy_clean()
            for elt in self:
                if callable(kw[key]):
                    if kw[key](elt) == True:
                        obj.append(elt)
                else:
                    if getattr(elt, key) == kw[key]:
                        obj.append(elt)
            del kw[key]
            return obj.filter(**kw)
        return self


class Email(GenericModel):

    def __init__(self, cursor, **kw):
        GenericModel.__init__(self, **kw)
        self.cursor = cursor

    def _copy_clean(self):
        return Email(self.cursor)


class User(GenericModel):

    def __init__(self, cursor, **kw):
        GenericModel.__init__(self, **kw)
        self.cursor = cursor

    def __getattr__(self, id):
        if id == "emails":
            lst = GenericList()
            self.cursor.execute("select id, email from email where iduser=%s" % self.id)
            records = cursor.fetchall()
            for row in records:
                lst.append(Email(self.cursor, id = row[0], iduser=self.id, email=row[1]))
            #optimize query
            setattr(self, "emails", lst)
            return lst
        GenericModel.__getattr__(self, id)

class Users(GenericList):

    def __init__(self, cursor, load=True):
        GenericList.__init__(self)
        self.cursor = cursor
        if load:
            self.cursor.execute("select id, name from user")
            records = cursor.fetchall()
            for row in records:
                self.append(User(self.cursor, id = row[0], name=row[1]))

    def _copy_clean(self):
        return Users(self.cursor, load=False)


# GENERATE SGBD AND DATA
if os.path.exists("test.db"):
    os.remove("test.db")

sqliteConnection = sqlite3.connect('test.db')
sqlite_create_table_users = '''CREATE TABLE user (
                            id INTEGER PRIMARY KEY,
                            name TEXT NOT NULL);'''
sqlite_create_table_emails = '''CREATE TABLE email (
                            id INTEGER PRIMARY KEY,
                            iduser INTEGER,
                            email TEXT NOT NULL);'''

cursor = sqliteConnection.cursor()
print("Successfully Connected to SQLite")
cursor.execute(sqlite_create_table_users)
cursor.execute(sqlite_create_table_emails)
sqliteConnection.commit()
print("SQLite table created")
cursor.execute("insert into user(id, name) values (1, 'user1')")
cursor.execute("insert into user(id, name) values (2, 'user2')")
cursor.execute("insert into user(id, name) values (3, 'user3')")
cursor.execute("insert into email(id, iduser, email) values (1, 1, 'user1@mail1.com')")
cursor.execute("insert into email(id, iduser, email) values (2, 1, 'user1@mail2.com')")
cursor.execute("insert into email(id, iduser, email) values (3, 2, 'user2@mail1.com')")
sqliteConnection.commit()


# USE DATA
users = Users(cursor)
for user in users:
    print(user.id, user.name)

for user in users:
    print(user.id, user.name, user.emails.email)

a = users._copy_clean()

for user in users.filter(id=2):
    print(user.id, user.name, user.emails.email)

for user in users:
    print(user.id, user.name)
    for email in user.emails:
        print('    ', email.email)


cursor.close()
```

On doit encore pouvoir améliorer la création des listes avec des filtres pour éviter des faires des réquêtes non sélectives (sans where)
